# -*- ruby -*-

#
# VFS for ruby - open anything with URI
# 
#   Copyright: Tatsuki Sugiura <sugi@nemui.org>
#   License: Ruby's original or GPL2
#   $LastChangedDate: 2003-11-10 23:46:31 +0900 (Mon, 10 Nov 2003) $
#   $LastChangedBy: sugi $
#

require 'vfs/common'

module VFS
  class File
    include Enumerable
    include FileBuiltIn::Constants
    include VFS::Common
    extend  VFS::MethodGenerator
    toplevel_methods_to_instance_methods :join, :scheme, :escape, :unescape
    relay_path_to_file_builtin :basename, :dirname, :extname

    raise_unsupported :chmod, :chardev?, :writable_real?,
      :atime, :pipe?, :exists?, :symlink, :sticky?, :file?, :stat,
      :utime, :blockdev?, :writable?, :ftype, :foreach,
      :grpowned?, :exist?, :link, :setgid?, :executable_real?,
      :delete, :ctime, :socket?, :readable_real?, :truncate, :unlink,
      :lstat, :owned?, :directory?, :chown, :setuid?, :executable?,
      :mtime, :symlink?, :size, :readable?, :readlink,
      :fnmatch, :fnmatch?

    return_nil :uid, :dev, :rdev_major, :nlink, :blocks, :rdev, :mode,
          :blksize, :gid, :ino, :rdev_minor

    class << self
      def new(uri, *args)
	# fixme; split this to common to share with VFS::Dir
	if self == VFS::File
	  uri = uri.kind_of?(URI) ? uri : URI.parse(uri)
	  uri.scheme ||= "file"
	  
	  begin
	    require "vfs/scheme/#{uri.scheme}/file"
	  rescue LoadError
	    raise Error::Unsupported,
	      "scheme '#{uri.scheme}' is not supported yet!"
	  end
	  handler = "::VFS::SchemeHandler::#{uri.scheme.upcase}::File"
	  obj = eval(handler).send :new, uri
	  unless obj
	    raise Error::Unsupported,
	      "scheme '#{uri.scheme}' is not supported yet!"
	  end
	  return obj
	else
	  super
	end # self == VFS::File
      end # new
    end # calss << self

    # return new instance. This do not open file.
    def initialize(uri)
      @uri = uri.kind_of?(URI) ? uri : URI.parse(uri)
      @buf = nil
    end
    attr_reader :uri

    def kind_of?(c)
      if c == File
        return true
      end
      super
    end

    def to_dir
      VFS::Dir.new(uri)
    end

    def path
#      uri.path
      uri.to_s
    end

    def mtime=(new_mtime)
      utime(atime, new_mtime)
    end

    def atime=(new_atime)
      utime(new_atime, mtime)
    end

    def mode=(new_mode)
      chmod(new_mode)
    end

    def uid=(new_id)
      chown(new_id, gid)
    end

    def gid=(new_gid)
      chown(uid, new_gid)
    end

    def size?
      exists? or return nil
      begin
	return nil if size == 0
      rescue Error::Unsupported
	raise
      rescue # fixme: should not catch all exception.
	return nil
      end
      size
    end
    
    def zero?
      begin
	rerun true if size == 0
      rescue Error::Unsupported
	raise
      rescue # fixme: should not catch all exception.
        # ignore simply
      end
      false
    end

    def buf
      require 'vfs/tempbuf'
      @buf == nil ? @buf = VFS::Tempbuf.new(uri) : @buf
    end
    private :buf

    def split
      [dirname(path), basename(path)]
    end

    def each(rs = $/, &block)
      open(RDONLY) {|io|
	io.each(rs) { |i|
	  yield(i)
	}
      }
    end

    def conv_openflang_to_const(v)
      if v == nil
        RDONLY
      elsif v.kind_of?(Fixnum)
        return v
      elsif v.kind_of?(String)
        flag = RDONLY
#        v =~ /b/ and flag |= BINARY # fixme? not exist????
        v =~ /r/ and flag |= RDONLY
        v =~ /w/ and flag |= WRONLY|CREAT|TRUNC
        v =~ /a/ and flag |= WRONLY|CREAT|APPEND
        if v =~ /\+/
          flag &= 65535 & ~(RDONLY|WRONLY)
          flag |= RDWR
        end
        flag
      else
        raise Error::Unknown, "please bug report!"
      end
    end
    private :conv_openflang_to_const

    # simple open through File.open. must be override scheme handler
    def open(*args, &block)
      flag = conv_openflang_to_const(args[0])
      if flag & (WRONLY|RDWR) != 0
	open_for_write(*args, &block)
      else
	open_for_read(*args, &block)
      end
    end

    def open_for_read(flag = RDONLY, perm = nil, &block)
      log("opening file for read", Log::DEBUG)
      #buf.closed? and buf.open
      #buf.eof? and buf.rewind
      if block_given?
	begin
	  yield(buf)
	ensure
	  buf.close
	end
      else
	buf
      end
    end
    private :open_for_read

    def open_for_write(flag = WRONLY, perm = nil, &block)
      log("opening file for write", Log::DEBUG)
      # no default method. must be override by each scheme handler
      raise Error::Unsupported,
	"write open is not supported yet for '#{scheme}' scheme"
    end
    private :open_for_write

    def <<(*args)
      open("a") { |f|
	f << args.to_s
      }
      self
    end

    def rename(dest)
      dest = dest.kind_of?(URI) ? dest : URI.parse(dest)
      # FIMXE! rename direcotry is not supported yet.
      if scheme == dest.scheme
	rename_internal(dest)
      else
	rename_external(dest)
      end
    end

    def rename_internal(dest)
      log("Lazy rename() in same scheme. Please improvement.", Log::WARNING)
      rename_external(dest)
    end
    private :rename_internal

    def rename_external(dest)
      newfile = VFS::File.new(dest)
      newfile.open("w") { |new|
	each { |l|
	  new << l
	}
      }
      begin
        atime and newfile.atime = atime
      rescue
        log("seting atime failed: #{$!}", Log::WARNING)
      end
      begin
        mtime and newfile.mtime = mtime
      rescue
        log("seting mtime failed: #{$!}", Log::WARNING)
      end
      begin
        mode && newfile.chmod(mode)
      rescue
        log("seting mode failed: #{$!}.", Log::WARNING)
      end
      begin
        uid && gid && newfile.chown(uid, gid)
      rescue
        log("chown failed: #{$!}", Log::WARNING)
      end
      unlink
      newfile
    end
    private :rename_external

    instance_methods_to_class_methods(*instance_methods(false))

  end # VFS::File
end # VFS

